package org.jastadd.ragconnect.tests.singleList;

import org.jastadd.ragconnect.tests.AbstractMqttTest;
import org.jastadd.ragconnect.tests.TestUtils;
import org.junit.jupiter.api.Tag;
import org.junit.jupiter.api.Test;
import singleList.ast.MqttHandler;

import java.io.IOException;
import java.nio.file.Paths;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.TimeUnit;
import java.util.function.Function;

import static java.util.Collections.addAll;
import static org.assertj.core.util.Lists.newArrayList;
import static org.jastadd.ragconnect.tests.TestUtils.mqttUri;
import static org.jastadd.ragconnect.tests.TestUtils.testJaddContainReferenceToJackson;
import static org.jastadd.ragconnect.tests.singleList.AbstractSingleListTest.IntList.list;
import static org.junit.jupiter.api.Assertions.*;

/**
 * Base class for test cases "singleList manual" and "singleList incremental".
 *
 * @author rschoene - Initial contribution
 */
@Tag("List")
@Tag("SingleList")
public abstract class AbstractSingleListTest extends AbstractMqttTest {

  public interface TestWrapperJastAddList<T> extends Iterable<T> {
    int getNumChild();
  }
  public interface TestWrapperReceiverRoot {
    TestWrapperJastAddList<? extends TestWrapperA> getAList();
    TestWrapperJastAddList<? extends TestWrapperA> getAs();
    int getNumA();
    int getNumWithAddA();
    TestWrapperA getA(int index);

    TestWrapperJastAddList<? extends TestWrapperA> getWithAddAList();
    TestWrapperJastAddList<? extends TestWrapperA> getUsingWildcardAList();
    TestWrapperJastAddList<? extends TestWrapperA> getUsingWildcardWithAddAList();

    @SuppressWarnings("unused") boolean connectA(String mqttUri) throws IOException;
    boolean connectA(String mqttUri, int index) throws IOException;
    boolean connectUsingWildcardA(String mqttUri) throws IOException;
    @SuppressWarnings("unused") boolean connectUsingWildcardA(String mqttUri, int index) throws IOException;
    boolean connectWithAddA(String mqttUri) throws IOException;
    boolean connectUsingWildcardWithAddA(String mqttUri) throws IOException;
  }
  @SuppressWarnings("UnusedReturnValue")
  public interface TestWrapperSenderRoot {
    boolean connectA1(String mqttUri, boolean writeCurrentValue) throws IOException;
    boolean connectA2(String mqttUri, boolean writeCurrentValue) throws IOException;
    boolean connectA3(String mqttUri, boolean writeCurrentValue) throws IOException;
    boolean connectA4(String mqttUri, boolean writeCurrentValue) throws IOException;
    boolean connectInOutput(String mqttUri, boolean writeCurrentValue) throws IOException;

    TestWrapperSenderRoot setInput1(int input);
    TestWrapperSenderRoot setInput2(int input);
    TestWrapperSenderRoot setInput3(int input);
    TestWrapperSenderRoot setInput4(int input);
    TestWrapperSenderRoot setInOutput(int input);

    TestWrapperA getA1();
    TestWrapperA getA2();
    TestWrapperA getA3();
    TestWrapperA getA4();
  }
  public interface TestWrapperA {
    int getID();
  }

  AbstractSingleListTest(String shortName) {
    this.shortName = shortName;
  }

  protected static final String TOPIC_A_1 = "a/first";
  protected static final String TOPIC_A_2 = "a/second";
  protected static final String TOPIC_A_3 = "a/third";
  protected static final String TOPIC_A_4 = "a/fourth";
  protected static final String TOPIC_A_5_INOUT = "a/special";
  protected static final String TOPIC_A_WILDCARD = "a/#";

  protected TestWrapperSenderRoot senderRoot;
  protected TestWrapperReceiverRoot receiverRoot;
  protected ReceiverData data;

  private final String shortName;

  @Test
  public void checkJacksonReference() {
    testJaddContainReferenceToJackson(
        Paths.get("src", "test",
            "02-after-ragconnect", shortName, "RagConnect.jadd"), true);
  }

  @Override
  protected void setupReceiverAndConnect(boolean writeCurrentValue) throws IOException, InterruptedException {
    // late model initialization
    senderRoot.setInput1(0);
    senderRoot.setInput2(0);
    senderRoot.setInput3(0);
    senderRoot.setInput4(0);
    senderRoot.setInOutput(0);

    setupReceiverAndConnectPart();

    // connect. important: first receivers, then senders. to not miss initial value.

    // receive: explicit topic subscription
    assertTrue(receiverRoot.connectA(mqttUri(TOPIC_A_1), 0));
    assertTrue(receiverRoot.connectA(mqttUri(TOPIC_A_2), 1));
    assertTrue(receiverRoot.connectA(mqttUri(TOPIC_A_3), 2));
    assertTrue(receiverRoot.connectA(mqttUri(TOPIC_A_4), 3));
    assertTrue(receiverRoot.connectA(mqttUri(TOPIC_A_5_INOUT), 4));

    assertTrue(receiverRoot.connectWithAddA(mqttUri(TOPIC_A_1)));
    assertTrue(receiverRoot.connectWithAddA(mqttUri(TOPIC_A_2)));
    assertTrue(receiverRoot.connectWithAddA(mqttUri(TOPIC_A_3)));
    assertTrue(receiverRoot.connectWithAddA(mqttUri(TOPIC_A_4)));
    assertTrue(receiverRoot.connectWithAddA(mqttUri(TOPIC_A_5_INOUT)));

    // receive: wildcard subscription
    assertTrue(receiverRoot.connectUsingWildcardA(mqttUri(TOPIC_A_WILDCARD)));
    assertTrue(receiverRoot.connectUsingWildcardWithAddA(mqttUri(TOPIC_A_WILDCARD)));

    // send: explicit topics, wait between connections to ensure correct arrival at receiver
    MqttHandler checkArrivalHandler = new MqttHandler().dontSendWelcomeMessage().setHost(TestUtils.getMqttHost());
    Map<String, CountDownLatch> arrived = new HashMap<>() {{
      put(TOPIC_A_1, new CountDownLatch(1));
      put(TOPIC_A_2, new CountDownLatch(1));
      put(TOPIC_A_3, new CountDownLatch(1));
      put(TOPIC_A_4, new CountDownLatch(1));
    }};
    checkArrivalHandler.waitUntilReady(2, TimeUnit.SECONDS);
    checkArrivalHandler.newConnection("#", (topic, bytes) ->
        Optional.ofNullable(arrived.get(topic)).ifPresent(CountDownLatch::countDown));

    assertTrue(senderRoot.connectA4(mqttUri(TOPIC_A_4), writeCurrentValue));
    if (writeCurrentValue) {
      assertTrue(arrived.get(TOPIC_A_4).await(2, TimeUnit.SECONDS));
    }

    assertTrue(senderRoot.connectA3(mqttUri(TOPIC_A_3), writeCurrentValue));
    if (writeCurrentValue) {
      assertTrue(arrived.get(TOPIC_A_3).await(2, TimeUnit.SECONDS));
    }

    assertTrue(senderRoot.connectA2(mqttUri(TOPIC_A_2), writeCurrentValue));
    if (writeCurrentValue) {
      assertTrue(arrived.get(TOPIC_A_2).await(2, TimeUnit.SECONDS));
    }

    assertTrue(senderRoot.connectA1(mqttUri(TOPIC_A_1), writeCurrentValue));
    if (writeCurrentValue) {
      assertTrue(arrived.get(TOPIC_A_1).await(2, TimeUnit.SECONDS));
    }

    assertTrue(senderRoot.connectInOutput(mqttUri(TOPIC_A_5_INOUT), writeCurrentValue));
    // no need to wait here, because first "checkTree" will wait anyway
    checkArrivalHandler.close();
  }

  abstract protected void setupReceiverAndConnectPart() throws IOException;

  @Override
  protected void communicateSendInitialValue() throws InterruptedException {
    checkTree(5, list(1, 2, 3, 4, 0), list(4, 3, 2, 1, 0), // normal: (normal / wildcard)
                                     list(4, 3, 2, 1, 0), list(4, 3, 2, 1, 0)); // withAdd: (normal / wildcard)

    // A1 will be 2 (1+1, previously 1)
    setInput(1, 1);
    checkTree(6, list(2, 2, 3, 4, 0), list(4, 3, 2, 2, 0), // normal: (normal / wildcard)
                                     list(4, 3, 2, 1, 0, 2), list(4, 3, 2, 1, 0, 2)); // withAdd: (normal / wildcard)

    // A1 should stay at 2
    setInput(1, 1);
    checkTree(6, list(2, 2, 3, 4, 0), list(4, 3, 2, 2, 0), // normal: (normal / wildcard)
                                     list(4, 3, 2, 1, 0, 2), list(4, 3, 2, 1, 0, 2)); // withAdd: (normal / wildcard)

    // A1 will be 3 (2+1, previously 2)
    setInput(1, 2);
    checkTree(7, list(3, 2, 3, 4, 0), list(4, 3, 2, 3, 0), // normal: (normal / wildcard)
                                     list(4, 3, 2, 1, 0, 2, 3), list(4, 3, 2, 1, 0, 2, 3)); // withAdd: (normal / wildcard)

    // InOut will be 5 (previously 0)
    setInput(5, 5);
    checkTree(8, list(3, 2, 3, 4, 5), list(4, 3, 2, 3, 5), // normal: (normal / wildcard)
                                     list(4, 3, 2, 1, 0, 2, 3, 5), list(4, 3, 2, 1, 0, 2, 3, 5)); // withAdd: (normal / wildcard)

    // A3 will be 7 (4+3, previously 3)
    setInput(3, 4);
    checkTree(9, list(3, 2, 7, 4, 5), list(4, 7, 2, 3, 5), // normal: (normal / wildcard)
                                     list(4, 3, 2, 1, 0, 2, 3, 5, 7), list(4, 3, 2, 1, 0, 2, 3, 5, 7)); // withAdd: (normal / wildcard)
  }

  @Override
  protected void communicateOnlyUpdatedValue() throws InterruptedException {
    checkTree(0, list(0, 0, 0, 0, 0), list(), // normal
                                     list(), list()); // withAdd

    // A1 will be 2 (1+1, previously 1)
    setInput(1, 1);
    checkTree(1, list(2, 0, 0, 0, 0), list(2), // normal
                                     list(2), list(2)); // withAdd

    // A1 should stay at 2
    setInput(1, 1);
    checkTree(1, list(2, 0, 0, 0, 0), list(2), // normal
                                     list(2), list(2)); // withAdd

    // A1 will be 3 (2+1, previously 2)
    setInput(1, 2);
    checkTree(2, list(3, 0, 0, 0, 0), list(3), // normal
                                     list(2, 3), list(2, 3)); // withAdd

    // InOut will be 5 (previously 0)
    setInput(5, 5);
    checkTree(3, list(3, 0, 0, 0, 5), list(3, 5), // normal
                                     list(2, 3, 5), list(2, 3, 5)); // withAdd

    // A3 will be 7 (4+3, previously 3)
    setInput(3, 4);
    checkTree(4, list(3, 0, 7, 0, 5), list(3,5,7), // normal
                                     list(2, 3, 5, 7), list(2, 3, 5, 7)); // withAdd
  }

  protected void setInput(int index, int input) {
    int actualComputedValue;
    switch (index) {
      case 1: senderRoot.setInput1(input); actualComputedValue = senderRoot.getA1().getID(); break;
      case 2: senderRoot.setInput2(input); actualComputedValue = senderRoot.getA2().getID(); break;
      case 3: senderRoot.setInput3(input); actualComputedValue = senderRoot.getA3().getID(); break;
      case 4: senderRoot.setInput4(input); actualComputedValue = senderRoot.getA4().getID(); break;
      case 5: senderRoot.setInOutput(input); return;
      default: fail("Wrong index " + index); return;
    }
    assertEquals(input + index, actualComputedValue, "ID value of single A");
  }

  private void checkTree(int expectedTransmissions, IntList normalA, IntList usingWildcardA, IntList withAddA, IntList usingWildcardWithAddA) throws InterruptedException {
    TestUtils.waitForMqtt();
    assertEquals(expectedTransmissions, data.numberOfElements, "transmissions for any A");

    checkList(normalA.toList(), receiverRoot.getNumA(), receiverRoot::getA);
    checkList(normalA.toList(), receiverRoot.getAList());

    checkList(usingWildcardA.toList(), receiverRoot.getUsingWildcardAList());

    checkList(withAddA.toList(), receiverRoot.getWithAddAList());

    checkList(usingWildcardWithAddA.toList(), receiverRoot.getUsingWildcardWithAddAList());
  }

  private void checkList(List<Integer> expectedList, int numChildren, Function<Integer, TestWrapperA> getA) {
    assertEquals(expectedList.size(), numChildren, "same list size");
    for (int index = 0; index < expectedList.size(); index++) {
      TestWrapperA a = getA.apply(index);
      assertEquals(expectedList.get(index), a.getID(), "correct ID for A");
    }
  }

  private void checkList(List<Integer> expectedList, TestWrapperJastAddList<? extends TestWrapperA> actualList) {
    assertEquals(expectedList.size(), actualList.getNumChild(), "same list size");
    int index = 0;
    for (TestWrapperA a : actualList) {
      assertEquals(expectedList.get(index), a.getID(), "correct ID for A");
      index++;
    }
  }

  protected static class ReceiverData {
    int numberOfElements = 0;
  }

  protected static class IntList {
    private final List<Integer> integers = newArrayList();
    public IntList(Integer... values) {
      addAll(integers, values);
    }

    public List<Integer> toList() {
      return integers;
    }

    public static IntList list(Integer... values) {
      return new IntList(values);
    }
  }

}
